#include "stdafx.h"
#include "TGlobal.h"
#include "TAudioFile.h"
#include "TTools.h"
#include "TACMEncoder.h"
#include "TPref.h"

using namespace System;
using namespace System::Windows;

bool TACMEncoder::Init(TAudioFileProcess *context) {
	TEncoderBase::Init(context);
	memset(&m_fmt, 0, sizeof(m_fmt));
	m_fmt.wfx.wFormatTag = WAVE_FORMAT_MPEGLAYER3;
	m_enable = false;
	Load();
	return m_enable;
}

void TACMEncoder::Setup(HWND window) {
	MMRESULT mmr; 
	ACMFORMATCHOOSE afc; 
 	WAVEFORMATEX wfxEnum;
	memset(&afc, 0, sizeof(afc)); 
	afc.cbStruct    = sizeof(afc); 
	afc.fdwStyle    = ACMFORMATCHOOSE_STYLEF_INITTOWFXSTRUCT;
	afc.hwndOwner   = window;
	afc.pwfx        = &m_fmt.wfx;
	afc.cbwfx       = sizeof(m_fmt); 
	afc.pszTitle    = _T("Setup ACM Codec");
	memset(&wfxEnum, 0, sizeof(wfxEnum));
	wfxEnum.wFormatTag = WAVE_FORMAT_MPEGLAYER3; 
	afc.fdwEnum = ACM_FORMATENUMF_WFORMATTAG;
	afc.pwfxEnum = &wfxEnum;
 	mmr = acmFormatChoose(&afc);
	int x = 0;
	if (MMSYSERR_NOERROR == mmr) {
		m_enable = true;
		m_modify = true;
	}
	else if (ACMERR_CANCELED == mmr) {
	}
	else {
        m_enable = false;
        LPCTSTR msg = _T("Unknown Error");
        if (ACMERR_NOTPOSSIBLE == mmr) {
            msg = _T("ACMERR_NOTPOSSIBLE");
        }
        else if (MMSYSERR_INVALFLAG == mmr) {
            msg = _T("MMSYSERR_INVALFLAG");
        }
        else if (MMSYSERR_INVALHANDLE == mmr) {
            msg = _T("MMSYSERR_INVALHANDLE");
        }
        else if (MMSYSERR_INVALPARAM == mmr) {
            msg = _T("MMSYSERR_INVALPARAM");
        }
        else if (MMSYSERR_NODRIVER == mmr) {
            msg = _T("MMSYSERR_NODRIVER");
        }
        System::Windows::MessageBox::Show(gcnew String(msg), gcnew String(_T("TACMEncoder::Setup")));
    }
}

#pragma managed(push, off)

bool TACMEncoder::Encode(TAudioFile *af, TFileWriter *fOut, TAudioFilterReader *fIn, TAudioFileFormat *format) {
	if (IsEnable() == false)
		return false;
	if (format->m_waveFormat.wBitsPerSample != 16)
		return false;
	
	MPEGLAYER3WAVEFORMAT mp3FormatInstance = m_fmt;
	WAVEFORMATEX *mp3Format = &mp3FormatInstance.wfx;
//	mp3Format->nChannels = format->m_waveFormat.nChannels;
	mp3Format->nSamplesPerSec = format->m_waveFormat.nSamplesPerSec;
//	mp3Format->wBitsPerSample = waveFormat->wBitsPerSample;
	int sampleBytes = BitsToBytes(format->m_waveFormat.wBitsPerSample);

	DWORD bufferBytes = format->m_waveFormat.nSamplesPerSec * format->m_waveFormat.nChannels * 8;
	HACMSTREAM has = NULL;
	MMRESULT mmr; 
	mmr = acmStreamOpen(&has, NULL, &format->m_waveFormat, mp3Format, NULL, 0, 0, ACM_STREAMOPENF_NONREALTIME);
	if (mmr != 0) {
		return false;
	}
	ACMSTREAMHEADER ash;
	TByteBuffer srcBuffer(bufferBytes);
	TByteBuffer dstBuffer(bufferBytes);
	memset(&ash, 0, sizeof(ash));
	ash.cbStruct = sizeof(ash);
	ash.cbSrcLength = bufferBytes;
	ash.pbSrc = srcBuffer.m_buffer;
	ash.cbDstLength = bufferBytes;
	ash.pbDst = dstBuffer.m_buffer;
	mmr = acmStreamPrepareHeader(has, &ash, 0);
	if (mmr != 0) {
		return false;
	}

	DWORD bufferFrames = bufferBytes / (format->m_waveFormat.nChannels * sampleBytes);
	size_t rest = af->m_frameCount;
	long framePos = 0;
	while (rest > 0 && m_context->m_continue) {
		if (m_context->m_pause) {
			Sleep(100);
			continue;
		}
		int status = (int)(100.0 * ((double)af->m_frameCount - (double)rest) / (double)af->m_frameCount);
		if (af->m_status != status) {
			af->m_status = status;
			m_context->CallEncodeProgress(af);
		}
		memset(ash.pbSrc, 0, bufferBytes);
		memset(ash.pbDst, 0, bufferBytes);
		DWORD readFrames = rest;
		if (rest > bufferFrames)
			readFrames = bufferFrames;

		size_t p = 0;
		for (size_t i = 0; i < readFrames; i++) {
			for (int ch = 0; ch < format->m_waveFormat.nChannels; ch++) {
				long sample = (SHORT)fIn->GetSample(ch, framePos);
				SampleLongToRaw(sample, &srcBuffer.m_buffer[p], sampleBytes, false);
				p += sampleBytes;
			}
			framePos++;
		}

		DWORD readBytes = readFrames * format->m_waveFormat.nChannels * sampleBytes;
		p = 0;
		while (true) {
			ash.cbSrcLength = readBytes;
			ash.pbSrc = &srcBuffer.m_buffer[p];
			ash.cbSrcLengthUsed = 0;
			ash.cbDstLengthUsed = 0;
			mmr = acmStreamConvert(has, &ash, 0);
			fOut->Write(ash.pbDst, ash.cbDstLengthUsed);
			if (ash.cbSrcLengthUsed == readBytes)
				break;
			readBytes -= ash.cbSrcLengthUsed;
			p += ash.cbSrcLengthUsed;
		}
		rest -= readFrames;
	}
	if (m_context->m_continue) {
		af->m_status = 100;
		m_context->CallEncodeProgress(af);
	}
	return true;
}

#pragma managed(pop)

TACMEncoder::~TACMEncoder() {
	Save();
}

void TACMEncoder::Load() {
	struct _stat status;
	if (_wstat(GetSaveFilePath().c_str(), &status) != -1) {
		FILE *fp = _wfopen(GetSaveFilePath().c_str(), _T("rb"));
		if (fp != NULL) {
			fread(&m_fmt, sizeof(m_fmt), 1, fp);
			fclose(fp);
			m_enable = true;
		}
	}
	m_modify = false;
}

void TACMEncoder::Save() {
	if (m_modify) {
		FILE *fp = _wfopen(GetSaveFilePath().c_str(), _T("wb"));
		if (fp != NULL) {
			fwrite(&m_fmt, sizeof(m_fmt), 1, fp);
			fclose(fp);
		}
	}
}

std::wstring TACMEncoder::GetSaveFilePath() {
	std::wstring s = TStringUtil::MarshalString(Preferences::GetPrefDirPath());
	s += _T("MPEGLAYER3WAVEFORMAT.dat");
	return s;
}

