#ifndef _TAUDIOFILE_H_
#define _TAUDIOFILE_H_

#include "TTools.h"
#include "TACMEncoder.h"
//#include "TLameEncoder.h"
#include "TBladeEncoder.h"
#include "TAudioFileFormat.h"
#include "TGainFilter.h"
#include "TDeClickNoiseFilter.h"
#include "TDCOffsetFilter.h"
#include "TWaveFileWriter.h"
#include "TWaveFileReader.h"
#include "TFileIO.h"

class TAudioFile;
class TAudioFileProcess;

ref class TagOfAudioFile {
public:
	TAudioFile *m_af;
	TagOfAudioFile(TAudioFile *af) {
		m_af = af;
	}
};

#if USE_UNMANAGED
//#pragma managed(push, off)
#endif

#define	CODE_LATIN1		0
#define	CODE_UTF16		1

#define	TAG_VERSION_1_0		0
#define	TAG_VERSION_1_1		1
#define	TAG_VERSION_2_2		2
#define	TAG_VERSION_2_3		3
#define	TAG_VERSION_2_4		4

#define	TAG_V2_HEAD_FLAG_UNSYNC			0x80
#define	TAG_V2_HEAD_FLAG_EXTENDED		0x40
#define	TAG_V2_HEAD_FLAG_EXPERIMENTAL	0x20
#define	TAG_V2_HEAD_FLAG_FOOTER			0x10

struct ID3_V1 {
	char m_tag[3];
	char m_title[30];
	char m_artist[30];
	char m_album[30];
	char m_year[4];
	union {
		struct {
			char m_comment[30];
		}
		v10;
		struct {
			char m_comment[28];
			char m_null;
			char m_track;
		}
		v11;
	} u;
	char m_genre;
};

struct ID3_V2_HEAD {
	char m_tag[3];
	char m_major;
	char m_minor;
	char m_flags;
	char m_size[4];
};

struct ID3_V2_2_FRAME {
	char m_frameID[3];
	char m_size[3];
};

struct ID3_V2_2_FRAME_TEXT : public ID3_V2_2_FRAME {
	char m_encoding;
};

struct ID3_V2_2_FRAME_COMMENT : public ID3_V2_2_FRAME_TEXT {
	char m_language[3];
};

struct ID3_V2_3_FRAME {
	char m_frameID[4];
	char m_size[4];
	char m_flags[2];
};

struct ID3_V2_3_FRAME_TEXT : public ID3_V2_3_FRAME {
	char m_encoding;
};

struct ID3_V2_3_FRAME_COMMENT : public ID3_V2_3_FRAME_TEXT {
	char m_language[3];
};

class TProcessContext {
public:
	TAudioFileProcess *m_context;
	size_t m_bytesSum;
	int m_prevStatus;
	size_t m_bytesSumRead;
	TProcessContext(TAudioFileProcess *context, size_t bytesSum);
	bool IsContinue();
	void Progress(size_t addBytes);
};

class TPhaseGap
{
public:
	long m_pos;
	long m_gap;
	bool m_delete;
	TPhaseGap(long pos, long gap)
	{
		m_pos = pos;
		m_gap = gap;
		m_delete = FALSE;
	}
};

typedef ID3_V2_HEAD ID3_V2_FOOT;

class TAudioFile {
public:
	//TAudioFileProcess *m_context;
	std::wstring m_waveFilePath;
	std::wstring m_mp3FilePath;
	int m_status;
	std::wstring m_title;
	std::wstring m_artist;
	std::wstring m_album;
#define GAIN_CMD_GAIN				0
#define GAIN_CMD_LAUD				1
	int m_gainCmd;
	double m_gain;
	long m_maxSample;
	double m_maxDecibel;
	long m_maxSamplePos;
	long m_maxLaudSample;
	double m_maxLaudDecibel;
	TAudioFileFormat m_format;
	int m_sampleBytes;
	long m_frameCount;

	/* FixPhase */
	std::vector<TPhaseGap> m_phaseGapList;
	long m_phaseGapAverage;
	int m_phaseGapPercent;
	int m_phaseGapCheckStatus;
#define GAP_CHECK_STATUS_OK			1
#define GAP_CHECK_STATUS_NON		0
#define GAP_CHECK_STATUS_NG			(-1)
	int m_phaseCmd;
#define PHASE_CMD_NON				0
#define PHASE_CMD_AS_AVERAGE		1
#define PHASE_CMD_MOVE				2
	int m_phaseMove;

	/* Click Noise */
	TNoiseFramePosMultiCh m_noiseFramePos;
#define DECLICKNOISE_CMD_NON			0
#define DECLICKNOISE_CMD_DECLICK		1
#define DECLICKNOISE_CMD_DY1			2		// Differentiate
#define DECLICKNOISE_CMD_DY2			3
#define DECLICKNOISE_CMD_NOISE			4
	int m_deClickNoiseCmd;
	std::wstring m_ccnFilePath;
	std::wstring m_ccuFilePath;
	long m_noiseCheckStart;
	long m_noiseCheckEnd;

	double m_dcOffsetLeft;
	bool m_dcOffsetLeftDown;
	double m_dcOffsetRight;
	bool m_dcOffsetRightDown;
	double m_averageLeft;
	bool m_averageLeftPlus;
	double m_averageRight;
	bool m_averageRightPlus;

	TAudioFile();
	~TAudioFile();
	bool Setup(const std::wstring &waveFilePath);
	static bool CheckExtension(const std::wstring &path, const std::wstring &ext);
	static std::wstring ChangeExtension(const std::wstring &path, const std::wstring &newExt);
	static std::wstring ChangeDirectory(const std::wstring &path, const std::wstring &newDir);
	static std::wstring GetRoundDecibelString(double db);
	static double RoundDecibel(double db);
	std::wstring GetStatusString();
	std::wstring GetFileName(const std::wstring &path);
	std::wstring GetWaveFileName();
	std::wstring GetMp3FileName();
	std::wstring GetMaxDecibelString();
	std::wstring GetGainString();
	std::wstring GetPhaseGapAverageString();
	std::wstring GetMovePhaseString();
	std::wstring GetMaxPosString();
	std::wstring GetMaxLaudDecibelString();
	std::wstring GetClickNoiseCountString();
	std::wstring GetDeClickNoiseString();
	std::wstring GetAverageString();
	std::wstring GetMoveDCOffsetString();
	std::wstring GetHzString();
	std::wstring GetBitsString();
	std::wstring GetChString();
	std::wstring GetTimeString();
	int MakeV2Frame(BYTE *frame, char *frameID, int sz);
	int MakeV2TextFrame(BYTE *frame, char *frameID, const std::wstring &data);
	void SetV2TagSize(ID3_V2_HEAD *head, int sz);
	void SaveID3Tag(TFileWriter *fw);
	void MakeID3Data();
	bool CheckMaxDecibel(TProcessContext *context);
	bool CheckPhaseGap(TProcessContext *context);
	bool CheckClickNoise(TProcessContext *context);
	bool CheckPhaseGapLevel(long *buff, long count, double checkLevel);
	double CalcPhaseDiff(long *buff0, long *buff1, long count);
	static long GetPrefUnitCount();
	static long GetPrefEdgeCount();
	static double GetPrefCheckLevel();
	static long GetPrefGapHint();
	static bool GetPrefEnableOne();
	static bool GetPrefSaveLogOnCheckClickNoise();
	static bool GetPrefRepairOnCheckClickNoise();
	static int GetPrefClickNoiseAvrCount();
};

//typedef void (*CALLBACK_ENCODEPROGRESS)(TAudioFile *af);
//typedef void (*CALLBACK_ENCODEFINISHONE)(TAudioFile *af);

class TAudioFileProcess : public std::deque<TAudioFile *> {
protected:
	HANDLE m_thread;
	HANDLE m_threadPrev;
public:
	bool m_continue;
	bool m_pause;
	TGENERIC_CALLBACK EncodeProgress;
	TGENERIC_CALLBACK EncodeFinishOne;
	TGENERIC_CALLBACK CommonProgress;
	TGENERIC_CALLBACK CommonStartOne;
	TGENERIC_CALLBACK CommonFinishOne;
	TGENERIC_CALLBACK CommonFinish;
	TACMEncoder m_acmEncoder;
	TBladeEncoder m_lameEncoder;
	//TLameEncoder m_lameEncoder;
	TEncoderBase *m_selectedEncoder;
	std::vector<TEncoderBase *> m_encoders;
	int m_priority;
	TCriticalSection m_csCallback;

	// Xbhp^B
	int m_threadCommand;
#define THR_CMD_ENCODE					0
#define THR_CMD_SAVE_FILES_TO			1
#define THR_CMD_SAVE_MERGE_FILE			2
#define THR_CMD_CHECK_MAX_DECIBEL		3
#define THR_CMD_CHECK_PHASE_GAP			4
#define THR_CMD_CHECK_CLICK_NOISE		5
#define THR_CMD_SAVE_FILE_AS			6
	std::wstring m_saveFilesToDirPath;
	std::wstring m_saveMergeFilePath;
	std::wstring m_saveFileAsPath;
	TAudioFile *m_saveFileAsAf;

	TAudioFileProcess();
	~TAudioFileProcess();
	void RemoveDeleteAll();
	void ClearAllStatus();
	void StartThread();
	void StartEncode();
	void Stop();
	void PauseEncode(bool pause);
	bool IsWorking();
	void Join();
	int GetThreadCommand();
	void SetPriority(int priority);
	static unsigned CALLBACK ThreadMain(void *param);
	bool Encode();
	bool GetAudioFileFormat(LPCTSTR filePath, WAVEFORMATEX *waveFormat, long *startPos);
	bool GetWaveFormat(LPCTSTR filePath, WAVEFORMATEX *waveFormat, long *startPos);
	void CallEncodeProgress(TAudioFile *af);
	void CallEncodeFinishOne(TAudioFile *af);
	void CallCommonProgress(int percent);
	void CallCommonStartOne(TAudioFile *af);
	void CallCommonFinishOne(TAudioFile *af);
	void CallCommonFinish();
	void StartSaveFilesTo(const std::wstring &dirPath);
	void StartSaveMergeFile(const std::wstring &savePath);
	void StartSaveFileAs(TAudioFile *af, const std::wstring &savePath);
	void StartCheckMaxDecibel();
	void StartCheckPhaseGap();
	void StartCheckClickNoise();
	bool SaveFilesTo(const std::wstring &dirPath);
	bool SaveFileAs(TAudioFile *af, const std::wstring &savePath);
	bool SaveMergeFile(const std::wstring &savePath);
	size_t CalcAllDataBytes();
	size_t CalcAllDataFrames();
	bool CheckMaxDecibel();
	bool CheckPhaseGap();
	bool CheckClickNoise();
};

class TAudioFilterReader {
public:
	TAudioFile *m_af;
	TWaveFileReader *m_wfr;
	TGainFilter *m_gainFilter;
	TDeClickNoiseFilter *m_deClickNoiseFilter;
	TDifferentiateFilter *m_differentiateFilter;
	TDCOffsetFilter *m_dcOffsetFilter;
	int m_phaseMove;
	TAudioFilterReader();
	~TAudioFilterReader();
	bool Open(TAudioFile *af);
	long GetSample(int ch, long framePos);
};

#if USE_UNMANAGED
//#pragma managed(pop)
#endif

#endif	// _AUDIOFILE_H_
