#include "stdafx.h"
#include "TAudioFileFormat.h"

TAudioFileFormat::TAudioFileFormat() {
	memset(&m_waveFormat, 0, sizeof(WAVEFORMATEX));
	m_bigEndian = false;
	m_dataPos = 0;
}

bool TAudioFileFormat::CheckAudioFileFormat(LPCTSTR filePath) {
	if (IsProbablyWave(filePath))
		return CheckWaveFormat(filePath);
	else if (IsProbablyAiff(filePath))
		return CheckAiffFormat(filePath);
	return false;
}

bool TAudioFileFormat::IsProbablyWave(LPCTSTR filePath) {
	FILE * fp = _wfopen(filePath, _T("rb"));
	if (fp == NULL)
		return false;
	BYTE data[16];
	fread(data, 16, 1, fp);
	fclose(fp);
	return (memcmp(&data[0], "RIFF", 4) == 0 &&
			memcmp(&data[8], "WAVE", 4) == 0 /*&&
			memcmp(&data[12], "fmt ", 4) == 0*/);
	// Windows Media PlayerŃbvWavet@CLIST`NĂB
	// 2010.09.06
}

bool TAudioFileFormat::CheckWaveFormat(LPCTSTR filePath) {
	MMIOINFO mmioinfo;
	memset(&mmioinfo, 0, sizeof(mmioinfo));
	HMMIO mmio = mmioOpen((LPTSTR)filePath, &mmioinfo, MMIO_READ);
	if (mmio == NULL)
		return false;

	MMRESULT mmr; 
	MMCKINFO ckWave;
	memset(&ckWave, 0, sizeof(ckWave));
	ckWave.fccType = ntohl('WAVE');
	mmr = mmioDescend(mmio, &ckWave, NULL, MMIO_FINDRIFF);
	if (!mmr == MMSYSERR_NOERROR) {
		mmr = mmioClose(mmio, 0);
		return false;
	}

	MMCKINFO ckFormat;
	memset(&ckFormat, 0, sizeof(ckFormat));
	ckFormat.ckid = ntohl('fmt ');
	mmr = mmioDescend(mmio, &ckFormat, &ckWave, MMIO_FINDCHUNK);
	if (!mmr == MMSYSERR_NOERROR) {
		mmr = mmioClose(mmio, 0);
		return false;
	}

	long allocSize = max(ckFormat.cksize, sizeof(WAVEFORMATEX));
	TByteBuffer waveFileFormatInstance(allocSize);
	WAVEFORMATEX *waveFileFormat = (WAVEFORMATEX *)waveFileFormatInstance.m_buffer;
	memset(waveFileFormat, 0, allocSize);
	mmr = mmioRead(mmio, (HPSTR)waveFileFormat, ckFormat.cksize);
	if (mmr == -1) {
		mmr = mmioClose(mmio, 0);
		return false;
	}

	MMCKINFO ckData;
	memset(&ckData, 0, sizeof(ckData));
	ckData.ckid = ntohl('data');
	mmr = mmioDescend(mmio, &ckData, &ckWave, MMIO_FINDCHUNK);
	if (!mmr == MMSYSERR_NOERROR) {
		mmr = mmioClose(mmio, 0);
		return false;
	}

	mmr = mmioClose(mmio, 0);
	memcpy(&m_waveFormat, waveFileFormat, sizeof(WAVEFORMATEX));
	m_dataPos = ckData.dwDataOffset;
	m_dataBytes = ckData.cksize;
	return true;
}

bool TAudioFileFormat::IsProbablyAiff(LPCTSTR filePath) {
	AIFF::ContainerChunk containerChunk;
	FILE *fp = _wfopen(filePath, _T("rb"));
	fread(&containerChunk, sizeof (AIFF::ContainerChunk), 1, fp);
	fclose(fp);
	return (ntohl(containerChunk.ckID) == AIFF::FORMID &&
		ntohl(containerChunk.formType) == AIFF::AIFFID);
}

short BitsToBytes(short bits) {
	short bytes = 0;
	for (;bits > 0;) {
		bytes += 1;
		bits -= 8;
	}
	return bytes;
}

short BitsChToBytes(short bits, short ch) {
	return ch * BitsToBytes(bits);
}

/*
#pragma managed(push, off)

double Extended80ToDouble(extended80 inExtended) {
	BYTE *bigExt = (byte *)&inExtended;
	BYTE littleExt[10];
	for (int i = 0; i < 10; i++) {
		littleExt[i] = bigExt[10 - i - 1];
	}
	double d;
	_asm {
		FLD tbyte ptr littleExt
		FWAIT;
		FSTP d;
		FWAIT;
	}
	return d;
}

#pragma managed(pop)
*/

bool TAudioFileFormat::CheckAiffFormat(LPCTSTR filePath) {
	AIFF::ContainerChunk containerChunk;
	AIFF::CommonChunk commonChunk;
	AIFF::SoundDataChunk soundDataChunk;

	FILE *fp = _wfopen(filePath, _T("rb"));
	fread(&containerChunk, sizeof (AIFF::ContainerChunk), 1, fp);
	if (ntohl(containerChunk.ckID) != AIFF::FORMID || ntohl(containerChunk.formType) != AIFF::AIFFID) {
		fclose(fp);
		return false;
	}

	size_t rest = ntohl(containerChunk.ckSize) - sizeof (AIFF::ID);

	bool bReadCommon = false;
	bool bReadSoundData = false;

	for (; rest >= sizeof (AIFF::ChunkHeader);) {
		if (bReadCommon && bReadSoundData)
			break;
		AIFF::ChunkHeader head;
		fread(&head, sizeof (AIFF::ChunkHeader), 1, fp);
		if (head.ckID == 0 || head.ckSize == 0) {
			break;
		}
		else if (ntohl(head.ckID) == AIFF::CommonID) {
			memcpy(&commonChunk, &head, sizeof (AIFF::ChunkHeader));
			fread(((char *)&commonChunk) + sizeof (AIFF::ChunkHeader), sizeof (AIFF::CommonChunk) - sizeof (AIFF::ChunkHeader), 1, fp);
			bReadCommon = true;
		}
		else if (ntohl(head.ckID) == AIFF::SoundDataID) {
			memcpy(&soundDataChunk, &head, sizeof (AIFF::ChunkHeader));
			fread(((char *)&soundDataChunk) + sizeof (AIFF::ChunkHeader), sizeof (AIFF::SoundDataChunk) - sizeof (AIFF::ChunkHeader), 1, fp);
			m_dataPos = ftell(fp) + ntohl(soundDataChunk.offset);
			bReadSoundData = true;
		}
		else {
			fseek(fp, ntohl(head.ckSize), SEEK_CUR);
		}
		rest -= (sizeof (AIFF::ChunkHeader) + ntohl(head.ckSize));
	}
	fclose(fp);
	if (bReadCommon == false || bReadSoundData == false)
		return false;

	if (ntohl(commonChunk.numSampleFrames) != 0)
		m_dataBytes = ntohl(commonChunk.numSampleFrames) * BitsChToBytes(ntohs(commonChunk.sampleSize), ntohs(commonChunk.numChannels));
	else
		m_dataBytes = ntohl(soundDataChunk.ckSize) - ntohl(soundDataChunk.offset) - (sizeof (AIFF::SoundDataChunk) - sizeof (AIFF::ChunkHeader));

	m_waveFormat.wFormatTag = WAVE_FORMAT_PCM;
    m_waveFormat.nChannels = ntohs(commonChunk.numChannels);
//    m_waveFormat.nSamplesPerSec = (DWORD)Extended80ToDouble(commonChunk.sampleRate); 
    m_waveFormat.nSamplesPerSec = (DWORD)ConvertFromIeeeExtended(&commonChunk.sampleRate); 
    m_waveFormat.nAvgBytesPerSec = 
		m_waveFormat.nSamplesPerSec * BitsChToBytes(ntohs(commonChunk.sampleSize), ntohs(commonChunk.numChannels)); 
    m_waveFormat.nBlockAlign = BitsChToBytes(ntohs(commonChunk.sampleSize), ntohs(commonChunk.numChannels)); 
    m_waveFormat.wBitsPerSample = ntohs(commonChunk.sampleSize);
    m_waveFormat.cbSize = 0; 

	m_bigEndian = true;
	return true;
}

bool TAudioFileFormat::IsCompatible(const TAudioFileFormat *format) {
	return (m_waveFormat.wFormatTag == format->m_waveFormat.wFormatTag &&
	m_waveFormat.nChannels == format->m_waveFormat.nChannels &&
	m_waveFormat.nSamplesPerSec == format->m_waveFormat.nSamplesPerSec &&
	m_waveFormat.wBitsPerSample == format->m_waveFormat.wBitsPerSample);
}

long TAudioFileFormat::CalcTimeLong() {
	return m_dataBytes / (BitsChToBytes(m_waveFormat.wBitsPerSample, m_waveFormat.nChannels) * m_waveFormat.nSamplesPerSec);
}

#pragma managed(push, off)

// 삵Ă}ṼA[LeN`ɋˑR[hB
// GfBAɈˑȂR[h͏͂Aʓ|Ȃ̂ŁA܂B
long SampleRawToLong(const BYTE *raw, int bytes, bool bBigEndian) {
	union {
		BYTE bx[4];
		long ix;
	} data;
	data.ix = 0;
	for (int i = 0; i < bytes; i++) {
		data.bx[4 - bytes + i] = raw[bBigEndian ? bytes - i - 1 : i];
	}
	long x = (data.ix >> (32 - bytes * 8));
	return x;
}

void SampleLongToRaw(long sample, BYTE *raw, int bytes, bool bBigEndian) {
	union {
		BYTE bx[4];
		long ix;
	} data;
	data.ix = sample;
	for (int i = 0; i < bytes; i++) {
		raw[bBigEndian ? bytes - i - 1 : i] = data.bx[i];
	}
}

#pragma managed(pop)

