#ifndef _TAUDIOFILEFORMAT_H_
#define _TAUDIOFILEFORMAT_H_

#include "TTools.h"
#include "ieee.h"

namespace AIFF {

typedef unsigned char                   UInt8;
typedef signed char                     SInt8;
//typedef unsigned short                  UInt16;
typedef signed short                    SInt16;
typedef unsigned long                   UInt32;
typedef signed long                     SInt32;

enum {
  AIFFID                        = ('AIFF'),
  AIFCID                        = ('AIFC'),
  FormatVersionID               = ('FVER'),
  CommonID                      = ('COMM'),
  FORMID                        = ('FORM'),
  SoundDataID                   = ('SSND'),
  MarkerID                      = ('MARK'),
  InstrumentID                  = ('INST'),
  MIDIDataID                    = ('MIDI'),
  AudioRecordingID              = ('AESD'),
  ApplicationSpecificID         = ('APPL'),
  CommentID                     = ('COMT'),
  NameID                        = ('NAME'),
  AuthorID                      = ('AUTH'),
  CopyrightID                   = ('(c) '),
  AnnotationID                  = ('ANNO')
};

#pragma pack(push, 2)

//struct Float80 {
//    SInt16  exp;
//    unsigned short  man[4];
//};
//typedef struct Float80 Float80;
//typedef Float80                         extended80;

typedef unsigned long                   ID;
typedef short                           MarkerIdType;
struct ChunkHeader {
  ID                  ckID;
  long                ckSize;
};
typedef struct ChunkHeader              ChunkHeader;
struct ContainerChunk {
  ID                  ckID;
  long                ckSize;
  ID                  formType;
};
typedef struct ContainerChunk           ContainerChunk;
struct FormatVersionChunk {
  ID                  ckID;
  long                ckSize;
  unsigned long       timestamp;
};
typedef struct FormatVersionChunk       FormatVersionChunk;
typedef FormatVersionChunk *            FormatVersionChunkPtr;
struct CommonChunk {
  ID                  ckID;
  long                ckSize;
  short               numChannels;
  unsigned long       numSampleFrames;
  short               sampleSize;
  IEEE::extended80          sampleRate;
};
typedef struct CommonChunk              CommonChunk;
typedef CommonChunk *                   CommonChunkPtr;
struct ExtCommonChunk {
  ID                  ckID;
  long                ckSize;
  short               numChannels;
  unsigned long       numSampleFrames;
  short               sampleSize;
  IEEE::extended80          sampleRate;
  ID                  compressionType;
  char                compressionName[1];     /* variable length array, Pascal string */
};
typedef struct ExtCommonChunk           ExtCommonChunk;
typedef ExtCommonChunk *                ExtCommonChunkPtr;
struct SoundDataChunk {
  ID                  ckID;
  long                ckSize;
  unsigned long       offset;
  unsigned long       blockSize;
};
typedef struct SoundDataChunk           SoundDataChunk;
typedef SoundDataChunk *                SoundDataChunkPtr;

#pragma pack(pop)

};	// namespace

class TAudioFileFormat {
protected:
	bool IsProbablyWave(LPCTSTR filePath);
	bool IsProbablyAiff(LPCTSTR filePath);
	bool CheckWaveFormat(LPCTSTR filePath);
	bool CheckAiffFormat(LPCTSTR filePath);
public:
	WAVEFORMATEX m_waveFormat;
	bool m_bigEndian;
	DWORD m_dataPos;
	DWORD m_dataBytes;
	TAudioFileFormat();
	bool CheckAudioFileFormat(LPCTSTR filePath);
	bool IsCompatible(const TAudioFileFormat *format);
	long CalcTimeLong();
};

short BitsToBytes(short bits);
short BitsChToBytes(short bits, short ch);
long SampleRawToLong(const BYTE *raw, int bytes, bool bBigEndian);
void SampleLongToRaw(long sample, BYTE *raw, int bytes, bool bBigEndian);

#endif	// _AUDIOFILEFORMAT_H_
