#include "stdafx.h"
#include "TGlobal.h"
#include "TDeClickNoiseFilter.h"
#include "InterpolateAudio.h"
#include "TFileIO.h"

TClickNoiseChecker::TClickNoiseChecker(TWaveFileReader *wfr, int ch, long frameCount, long st, long en) {
	Setup(wfr, ch, frameCount, st, en);
}

void TClickNoiseChecker::Setup(TWaveFileReader *wfr, int ch, long frameCount, long st, long en) {
	m_wfr = wfr;
	m_ch = ch;
	m_frameCount = frameCount;
	m_samplePrev = 0;
	m_delta1Prev = 0;
	//	
	m_gf.Setup(m_wfr->GetFormat()->m_waveFormat.wBitsPerSample);
	m_clickNoiseThrFix = Prefs::data->clickNoiseThrFix;
	m_clickNoiseThrAvrMag1 = Prefs::data->clickNoiseThrAvrMag1;
	m_clickNoiseThrAvrMag2 = Prefs::data->clickNoiseThrAvrMag2;
	m_clickNoiseCvgThrAvrMag = Prefs::data->clickNoiseCvgThrAvrMag;
	m_clickNoiseAvrCount = Prefs::data->clickNoiseAvrCount;
	m_clickNoiseMaxDistance = Prefs::data->clickNoiseMaxDistance;
	m_clickNoiseLeft = Prefs::data->clickNoiseLeft;
	m_clickNoiseRight = Prefs::data->clickNoiseRight;
	//
	m_averOfDelta2.Setup(m_clickNoiseAvrCount);	// 20
	m_averOfSample.Setup(m_clickNoiseAvrCount);

	m_startFramePos = st;
	m_endFramePos = en;
}

TDeClickNoiseFilter::TDeClickNoiseFilter(TWaveFileReader *wfr, const TNoiseFramePosMultiCh *noiseFramePos, const std::wstring &ccnFilePath, const std::wstring &ccuFilePath, bool deClick) {
	m_wfr = wfr;
	m_deClick = deClick;
	int ch = m_wfr->GetFormat()->m_waveFormat.nChannels;
	m_noiseFramePos.Setup(ch);
	if (noiseFramePos->size() == ch) {
		m_noiseFramePos.Merge(noiseFramePos);
	}
	else {
		m_noiseFramePos.Load(ccnFilePath, ch);
	}
	TNoiseFramePosMultiCh userPos;
	if (userPos.Load(ccuFilePath, ch)) {
		m_noiseFramePos.Merge(&userPos);
	}
	for (int i = 0; i < ch; i++) {
		m_repairNoise.push_back(new TRepairNoise());
	}
	m_index = new int[ch];
	for (int i = 0; i < ch; i++) {
		m_index[i] = 0;
	}
}

TDeClickNoiseFilter::~TDeClickNoiseFilter() {
	for (int i = 0; i < (int)m_repairNoise.size(); i++)
		delete m_repairNoise[i];
	delete [] m_index;
}

#pragma managed(push, off)

void TNoiseFramePosMultiCh::Setup(int ch) {
	clear();
	for (int i = 0; i < ch; i++) {
		push_back(std::vector<TNoiseFramePos>());
	}
}

long TNoiseFramePosMultiCh::GetFramesSum(int ch) {
	long sum = 0;
	for (int i = 0; i < (int)at(ch).size(); i++) {
		sum += at(ch)[i].GetLength();
	}
	return sum;
}

long TNoiseFramePosMultiCh::GetFramesSum() {
	long sum = 0;
	for (int ch = 0; ch < (int)size(); ch++) {
		sum += GetFramesSum(ch);
	}
	return sum;
}

void TNoiseFramePosMultiCh::Sort() {
	for (int ch = 0; ch < (int)size(); ch++)
		std::stable_sort(at(ch).begin(), at(ch).end(), Less());
}

void TNoiseFramePosMultiCh::Merge(const TNoiseFramePosMultiCh *p) {
	for (int ch = 0; ch < (int)size(); ch++) {
		for (int i = 0; i < (int)p->at(ch).size(); i++) {
			at(ch).push_back(p->at(ch)[i]);
		}
	}
	Sort();
}

bool TNoiseFramePosMultiCh::Save(const std::wstring &path) {
	TNoiseFramePosMultiCh p;
	p.Setup(size());
	p.Merge(this);
	TFileWriter fw(path);
	if (!fw.IsOK())
		return false;
	for (int ch = 0; ch < (int)p.size(); ch++) {
		for (int i = 0; i < (int)p[ch].size(); i++) {
			char buffer[80];
			sprintf(buffer, "%d %d %d\r\n", ch, p[ch][i].m_start, p[ch][i].m_end);
			fw.Write(buffer, strlen(buffer));
		}
	}
	return true;
}

bool TNoiseFramePosMultiCh::Load(const std::wstring &path, int ch) {
	if (!TFileReader::IsExist(path))
		return false;
	TTextFileReader tfr(path);
	if (!tfr.IsOK())
		return false;
	Setup(ch);
	while (true) {
		std::string line;
		if (!tfr.ReadLine(&line))
			break;
		if (line.length() == 0)
			continue;
		if (line[0] == '#')
			continue;
		std::vector<std::string> svalue;
		if (TStringUtil::Split(line, " \t", &svalue) < 3)
			continue;
		std::vector<int> value;
		for (int i = 0; i < 3; i++) {
			if (TStringUtil::IsDigit(svalue[i]))
				value.push_back(TStringUtil::ToInteger(svalue[i]));
		}
		if (value.size() != 3)
			continue;
		long c = value[0];
		if (value[0] >= ch)
			continue;
		at(value[0]).push_back(TNoiseFramePos(value[1], value[2]));
	}
	Sort();
	return true;
}

long TClickNoiseChecker::CorrectFramePos(long f) {
	if (f < 0)
		f = 0;
	if (f >= m_frameCount)
		f = m_frameCount - 1;
	return f;
}

long TClickNoiseChecker::SearchStart(long framePos, long cnv, long maxDistance) {
	long fPos = framePos;
	std::vector<long> d2s;
	int checkCount = m_clickNoiseAvrCount;	// 30
	bool hit = false;
	for (long i = 0; i < maxDistance && hit == false; i++) {
		fPos--;
		long sample0 = m_wfr->GetSample(m_ch, fPos - 2);
		long sample1 = m_wfr->GetSample(m_ch, fPos - 1);
		long sample2 = m_wfr->GetSample(m_ch, fPos);
		long d1_0 = sample1 - sample0;
		long d1_1 = sample2 - sample1;
		long d2 = d1_1 - d1_0;
		d2s.insert(d2s.begin(), abs(d2));
		while ((int)d2s.size() > checkCount) {
			d2s.pop_back();
		}
		if ((int)d2s.size() == checkCount) {
			hit = true;
			for (int i = 0; i < (int)d2s.size(); i++) {
				if (d2s[i] > cnv)
					hit = false;
			}
		}
		if (hit)
			return CorrectFramePos(fPos + checkCount - 1 - 1);	// 1]ɖ߂B
	}
	return -1;
}

long TClickNoiseChecker::SearchEnd(long framePos, long cnv, long maxDistance) {
	long fPos = framePos;
	std::vector<long> d2s;
	int checkCount = m_clickNoiseAvrCount;	// 30
	bool hit = false;
	for (int i = 0; i < maxDistance && hit == false; i++) {
		fPos++;
		long sample0 = m_wfr->GetSample(m_ch, fPos - 2);
		long sample1 = m_wfr->GetSample(m_ch, fPos - 1);
		long sample2 = m_wfr->GetSample(m_ch, fPos);
		long d1_0 = sample1 - sample0;
		long d1_1 = sample2 - sample1;
		long d2 = d1_1 - d1_0;
		d2s.insert(d2s.begin(), abs(d2));
		while ((int)d2s.size() > checkCount) {
			d2s.pop_back();
		}
		if ((int)d2s.size() == checkCount) {
			hit = true;
			for (int i = 0; i < (int)d2s.size(); i++) {
				if (d2s[i] > cnv)
					hit = false;
			}
		}
		if (hit)
			return CorrectFramePos(fPos - checkCount + 1 + 2);	// 2]ɐi߂B
	}
	return -1;
}

bool TClickNoiseChecker::CheckNoise(long framePos, long sample, TNoiseFramePosMultiCh *noiseFramePos) {
	bool result = false;
	long d1 = sample - m_samplePrev;
	long d2 = d1 - m_delta1Prev;
	m_samplePrev = sample;
	m_delta1Prev = d1;
	long ad2 = abs(d2);

//////////////////////////////////
//if (m_ch == 1 && framePos == 1162276) {
//int x = 0;
//}
//////////////////////////////////

	if (noiseFramePos->size() == 2 && ((!m_clickNoiseLeft && m_ch == 0) || (!m_clickNoiseRight && m_ch == 1))) {
	}
	else if (framePos < 3) {
	}
	else if (framePos < m_startFramePos) {
	}
	else if (m_endFramePos != -1 && framePos > m_endFramePos) {
	}
	else {
		int sz = (int)noiseFramePos->at(m_ch).size();
		for (int i = 0; i < (int)sz; i++) {
			int index = sz - 1 - i;
			if (noiseFramePos->at(m_ch)[index].m_start <= framePos && framePos <= noiseFramePos->at(m_ch)[index].m_end) {
				result = true;
				break;
			}
			if (noiseFramePos->at(m_ch)[index].m_end < framePos) {
				break;
			}
		}
		if (!result && m_gf.SampleToDecibel(ad2) > m_clickNoiseThrFix) {	// -50.0
			long averOfSample = m_averOfSample.GetAverage();
			if (ad2 * m_clickNoiseThrAvrMag2 > averOfSample) {	// 10
				long avr = m_averOfDelta2.GetAverage();
				if (ad2 > avr * m_clickNoiseThrAvrMag1) {	// 12
					long st = SearchStart(framePos, avr * m_clickNoiseCvgThrAvrMag, m_clickNoiseMaxDistance);	// 3
					long en = SearchEnd(framePos, avr * m_clickNoiseCvgThrAvrMag, m_clickNoiseMaxDistance - (framePos - st));	// 3
					if (st >= 0 && en >= 0) {
						//if (((en - st + 1) & 1))
						//	en++;	// ɂB
						TNoiseFramePos np(st, en);
						noiseFramePos->at(m_ch).push_back(np);
						result = true;
					}
				}
			}
		}
	}
	m_averOfDelta2.AddAbs(ad2);
	m_averOfSample.Add(sample);

	return result;
}

long TDeClickNoiseFilter::DeClickNoise(int ch, long framePos, long sample) {
	if (m_deClick) {
		if (m_repairNoise[ch]->m_repairBuffer != NULL && 
			m_repairNoise[ch]->m_currentNoise.m_start <= framePos && framePos <= m_repairNoise[ch]->m_currentNoise.m_end) {
		}
		else {
			int p = -1;
			while (m_index[ch] < (int)m_noiseFramePos[ch].size()) {
				if (m_noiseFramePos[ch][m_index[ch]].m_start <= framePos && framePos <= m_noiseFramePos[ch][m_index[ch]].m_end) {
					p = m_index[ch];
					break;
				}
				if (m_noiseFramePos[ch][m_index[ch]].m_start > framePos)
					break;
				m_index[ch] += 1;
			}
			if (p != -1) {
				m_repairNoise[ch]->m_currentNoise = m_noiseFramePos[ch][p];
				long nlen = m_repairNoise[ch]->m_currentNoise.m_end - m_repairNoise[ch]->m_currentNoise.m_start + 1;
				m_repairNoise[ch]->m_offset = max(nlen * 2, 128);
				long top = m_repairNoise[ch]->m_currentNoise.m_start - m_repairNoise[ch]->m_offset;
				long len = m_repairNoise[ch]->m_offset * 2 + nlen;
				if (m_repairNoise[ch]->m_repairBuffer != NULL)
					delete [] m_repairNoise[ch]->m_repairBuffer;
				m_repairNoise[ch]->m_repairBuffer = new float[len];
				for (long i = 0; i < len; i++) {
					m_repairNoise[ch]->m_repairBuffer[i] = (float)m_wfr->GetSample(ch, top + i);
				}
				Audacity::InterpolateAudio(m_repairNoise[ch]->m_repairBuffer, len, m_repairNoise[ch]->m_offset, nlen);
			}
		}
		if (m_repairNoise[ch]->m_repairBuffer != NULL &&
			m_repairNoise[ch]->m_currentNoise.m_start <= framePos && framePos <= m_repairNoise[ch]->m_currentNoise.m_end) {
				sample = (long)m_repairNoise[ch]->m_repairBuffer[framePos - m_repairNoise[ch]->m_currentNoise.m_start + m_repairNoise[ch]->m_offset];
		}
	}
	else {
		int p = -1;
		while (m_index[ch] < (int)m_noiseFramePos[ch].size()) {
			if (m_noiseFramePos[ch][m_index[ch]].m_start <= framePos && framePos <= m_noiseFramePos[ch][m_index[ch]].m_end) {
				p = m_index[ch];
				break;
			}
			if (m_noiseFramePos[ch][m_index[ch]].m_start > framePos)
				break;
			m_index[ch] += 1;
		}
		if (p == -1)
			sample = 0;
	}
	return sample;
}

#pragma managed(pop)
