#include "stdafx.h"
#include "TWaveFileWriter.h"
#include "TAudioFileFormat.h"

#pragma managed(push, off)

#define WAVE_FILE_WRITER_BUFFER_BYTES		(1024 * 64)

TWaveFileWriter::TWaveFileWriter() {
	Init();
}

TWaveFileWriter::~TWaveFileWriter() {
	Close();
}

void TWaveFileWriter::Init() {
	m_hmmio = NULL;
	m_sampleBits = 0;
	m_sampleBytes = 0;
	m_count = 0;
	m_buffer = NULL;
}

bool TWaveFileWriter::Create(const std::wstring &filePath, PCMWAVEFORMAT *waveFormat, size_t dataBytes) {
	//WAVt@C̃I[v
	m_hmmio = mmioOpen((LPTSTR)filePath.c_str(), NULL, MMIO_CREATE | MMIO_WRITE);
	if (m_hmmio == NULL)
		return false;

	//"WAVE"`N
	memset(&m_ckWave, 0, sizeof(m_ckWave));
	m_ckWave.fccType = ntohl('WAVE');
	mmioCreateChunk(m_hmmio, &m_ckWave, MMIO_CREATERIFF);

	//"fmt "`N
	MMCKINFO ckFormat;
	memset(&ckFormat, 0, sizeof(ckFormat));
	ckFormat.ckid = ntohl('fmt ');
	ckFormat.cksize = sizeof(PCMWAVEFORMAT);
	mmioCreateChunk(m_hmmio, &ckFormat, 0);
	mmioWrite(m_hmmio, (HPSTR)waveFormat, ckFormat.cksize);
	mmioAscend(m_hmmio, &ckFormat, 0);

	//"data"`N
	m_ckData.ckid = ntohl('data');
	m_ckData.cksize = dataBytes;
	mmioCreateChunk(m_hmmio, &m_ckData, 0);

	m_sampleBits = waveFormat->wBitsPerSample;
	m_sampleBytes = BitsToBytes(m_sampleBits);
	m_count = 0;
	m_buffer = new BYTE[WAVE_FILE_WRITER_BUFFER_BYTES + 16];

	return true;
}

bool TWaveFileWriter::Write(long data) {
//if ((data & 0xFFFF)== 0xcdcd) {
//int x = 0;
//}
	if (m_hmmio == NULL)
		return false;
	BYTE raw[sizeof(long)];
	SampleLongToRaw(data, raw, m_sampleBytes, false);
	memcpy(&m_buffer[m_count], raw, m_sampleBytes);
	m_count += m_sampleBytes;
	if (m_count >= WAVE_FILE_WRITER_BUFFER_BYTES) {
		if (Flush() == false)
			return false;
	}
	return true;
}

bool TWaveFileWriter::Flush() {
	if (m_hmmio == NULL)
		return false;
	if (m_count > 0) {
		mmioWrite(m_hmmio, (HPSTR)m_buffer, m_count);
		m_count = 0;
	}
	return true;
}

//WAVt@C̃N[Y
void TWaveFileWriter::Close() {
	if (m_hmmio != NULL) {
		Flush();
		mmioAscend(m_hmmio, &m_ckData, 0);
		mmioAscend(m_hmmio, &m_ckWave, 0);
		mmioClose(m_hmmio, 0);
		delete [] m_buffer;
	}
	m_hmmio = NULL;
	m_buffer = NULL;
}

#pragma managed(pop)

